@@ -2,7 +2,7 @@
  * vim:ts=4:sw=4:expandtab
  *
  * i3bar - an xcb-based status- and ws-bar for i3
- * © 2010-2011 Axel Wagner and contributors (see also: LICENSE)
+ * © 2010-2012 Axel Wagner and contributors (see also: LICENSE)
  *
  * xcb.c: Communicating with X
  *
@@ -28,6 +28,7 @@
 struct xcb_color_strings_t {
     char *bar_fg;
     char *bar_bg;
+    char *sep_fg;
     char *active_ws_fg;
     char *active_ws_bg;
     char *active_ws_border;
@@ -69,13 +70,13 @@ void init_colors(const struct xcb_color_strings_t *colors);
  * Called once, before the program terminates.
  *
  */
-void clean_xcb();
+void clean_xcb(void);
 
 /*
  * Get the earlier requested atoms and save them in the prepared data-structure
  *
  */
-void get_atoms();
+void get_atoms(void);
 
 /*
  * Reparents all tray clients of the specified output to the root window. This
@@ -88,6 +89,15 @@ void get_atoms();
  */
 void kick_tray_clients(i3_output *output);
 
+/*
+ * We need to set the _NET_SYSTEM_TRAY_COLORS atom on the tray selection window
+ * to make GTK+ 3 applets with Symbolic Icons visible. If the colors are unset,
+ * they assume a light background.
+ * See also https://bugzilla.gnome.org/show_bug.cgi?id=679591
+ *
+ */
+void init_tray_colors(void);
+
 /*
  * Destroy the bar of the specified output
  *
@@ -98,24 +108,30 @@ void destroy_window(i3_output *output);
  * Reallocate the statusline-buffer
  *
  */
-void realloc_sl_buffer();
+void realloc_sl_buffer(void);
 
 /*
  * Reconfigure all bars and create new for newly activated outputs
  *
  */
-void reconfig_windows();
+void reconfig_windows(void);
 
 /*
  * Render the bars, with buttons and statusline
  *
  */
-void draw_bars();
+void draw_bars(bool force_unhide);
 
 /*
  * Redraw the bars, i.e. simply copy the buffer to the barwindow
  *
  */
-void redraw_bars();
+void redraw_bars(void);
+
+/*
+ * Set the current binding mode
+ *
+ */
+void set_current_mode(struct mode *mode);
 
 #endif
