@@ -4,22 +4,29 @@
  * i3 - an improved dynamic tiling window manager
  * © 2009-2011 Michael Stapelberg and contributors (see also: LICENSE)
  *
- * log.c: Setting of loglevels, logging functions.
+ * log.c: Logging functions.
  *
  */
-#ifndef _LOG_H
-#define _LOG_H
+#ifndef I3_LOG_H
+#define I3_LOG_H
 
 #include <stdarg.h>
 #include <stdbool.h>
 
+/* We will include libi3.h which define its own version of LOG, ELOG.
+ * We want *our* version, so we undef the libi3 one. */
+#if defined(LOG)
+#undef LOG
+#endif
+#if defined(ELOG)
+#undef ELOG
+#endif
 /** ##__VA_ARGS__ means: leave out __VA_ARGS__ completely if it is empty, that
    is, delete the preceding comma */
 #define LOG(fmt, ...) verboselog(fmt, ##__VA_ARGS__)
 #define ELOG(fmt, ...) errorlog("ERROR: " fmt, ##__VA_ARGS__)
-#define DLOG(fmt, ...) debuglog(LOGLEVEL, "%s:%s:%d - " fmt, __FILE__, __FUNCTION__, __LINE__, ##__VA_ARGS__)
+#define DLOG(fmt, ...) debuglog("%s:%s:%d - " fmt, I3__FILE__, __FUNCTION__, __LINE__, ##__VA_ARGS__)
 
-extern char *loglevels[];
 extern char *errorfilename;
 extern char *shmlogname;
 extern int shmlog_size;
@@ -32,10 +39,10 @@ extern int shmlog_size;
 void init_logging(void);
 
 /**
- * Enables the given loglevel.
+ * Set debug logging.
  *
  */
-void add_loglevel(const char *level);
+void set_debug_logging(const bool _debug_logging);
 
 /**
  * Set verbosity of i3. If verbose is set to true, informative messages will
@@ -47,29 +54,32 @@ void set_verbosity(bool _verbose);
 
 /**
  * Logs the given message to stdout while prefixing the current time to it,
- * but only if the corresponding debug loglevel was activated.
+ * but only if debug logging was activated.
  *
  */
-void debuglog(uint64_t lev, char *fmt, ...);
+void debuglog(char *fmt, ...)
+    __attribute__ ((format (printf, 1, 2)));
 
 /**
  * Logs the given message to stdout while prefixing the current time to it.
  *
  */
-void errorlog(char *fmt, ...);
+void errorlog(char *fmt, ...)
+    __attribute__ ((format (printf, 1, 2)));
 
 /**
  * Logs the given message to stdout while prefixing the current time to it,
  * but only if verbose mode is activated.
  *
  */
-void verboselog(char *fmt, ...);
+void verboselog(char *fmt, ...)
+    __attribute__ ((format (printf, 1, 2)));
 
 /**
- * Logs the given message to stdout while prefixing the current time to it.
- * This is to be called by LOG() which includes filename/linenumber
- *
+ * Deletes the unused log files. Useful if i3 exits immediately, eg.
+ * because --get-socketpath was called. We don't care for syscall
+ * failures. This function is invoked automatically when exiting.
  */
-void slog(char *fmt, va_list args);
+void purge_zerobyte_logfile(void);
 
 #endif
