@@ -9,8 +9,8 @@
  * (take your time to read it completely, it answers all questions).
  *
  */
-#ifndef _RANDR_H
-#define _RANDR_H
+#ifndef I3_RANDR_H
+#define I3_RANDR_H
 
 #include "data.h"
 #include <xcb/randr.h>
@@ -18,6 +18,11 @@
 TAILQ_HEAD(outputs_head, xoutput);
 extern struct outputs_head outputs;
 
+typedef enum {
+    CLOSEST_OUTPUT = 0,
+    FARTHEST_OUTPUT = 1
+} output_close_far_t;
+
 /**
  * We have just established a connection to the X server and need the initial
  * XRandR information to setup workspaces for each screen.
@@ -82,20 +87,29 @@ Output *get_output_by_name(const char *name);
  */
 Output *get_output_containing(int x, int y);
 
-/**
- * Gets the output which is the last one in the given direction, for example
- * the output on the most bottom when direction == D_DOWN, the output most
- * right when direction == D_RIGHT and so on.
- *
- * This function always returns a output.
- *
- */
-Output *get_output_most(direction_t direction, Output *current);
-
 /**
  * Gets the output which is the next one in the given direction.
  *
+ * If close_far == CLOSEST_OUTPUT, then the output next to the current one will
+ * selected.  If close_far == FARTHEST_OUTPUT, the output which is the last one
+ * in the given direction will be selected.
+ *
+ * NULL will be returned when no active outputs are present in the direction
+ * specified (note that ‘current’ counts as such an output).
+ *
  */
-Output *get_output_next(direction_t direction, Output *current);
+Output *get_output_next(direction_t direction, Output *current, output_close_far_t close_far);
+
+/**
+ * Like get_output_next with close_far == CLOSEST_OUTPUT, but wraps.
+ *
+ * For example if get_output_next(D_DOWN, x, FARTHEST_OUTPUT) = NULL, then
+ * get_output_next_wrap(D_DOWN, x) will return the topmost output.
+ *
+ * This function always returns a output: if no active outputs can be found,
+ * current itself is returned.
+ *
+ */
+Output *get_output_next_wrap(direction_t direction, Output *current);
 
 #endif
