@@ -7,8 +7,8 @@
  * xcursor.c: libXcursor support for themed cursors.
  *
  */
-#ifndef _XCURSOR_CURSOR_H
-#define _XCURSOR_CURSOR_H
+#ifndef I3_XCURSOR_CURSOR_H
+#define I3_XCURSOR_CURSOR_H
 
 #include <X11/Xlib.h>
 
@@ -16,7 +16,12 @@ enum xcursor_cursor_t {
     XCURSOR_CURSOR_POINTER = 0,
     XCURSOR_CURSOR_RESIZE_HORIZONTAL,
     XCURSOR_CURSOR_RESIZE_VERTICAL,
+    XCURSOR_CURSOR_TOP_LEFT_CORNER,
+    XCURSOR_CURSOR_TOP_RIGHT_CORNER,
+    XCURSOR_CURSOR_BOTTOM_LEFT_CORNER,
+    XCURSOR_CURSOR_BOTTOM_RIGHT_CORNER,
     XCURSOR_CURSOR_WATCH,
+    XCURSOR_CURSOR_MOVE,
     XCURSOR_CURSOR_MAX
 };
 
