@@ -1,3 +1,5 @@
+#undef I3__FILE__
+#define I3__FILE__ "load_layout.c"
 /*
  * vim:ts=4:sw=4:expandtab
  *
@@ -139,7 +141,7 @@ static int json_string(void *ctx, const unsigned char *val, size_t len) {
 #else
 static int json_string(void *ctx, const unsigned char *val, unsigned int len) {
 #endif
-    LOG("string: %.*s for key %s\n", len, val, last_key);
+    LOG("string: %.*s for key %s\n", (int)len, val, last_key);
     if (parsing_swallows) {
         /* TODO: the other swallowing keys */
         if (strcasecmp(last_key, "class") == 0) {
@@ -156,14 +158,19 @@ static int json_string(void *ctx, const unsigned char *val, unsigned int len) {
             memcpy(json_node->sticky_group, val, len);
             LOG("sticky_group of this container is %s\n", json_node->sticky_group);
         } else if (strcasecmp(last_key, "orientation") == 0) {
+            /* Upgrade path from older versions of i3 (doing an inplace restart
+             * to a newer version):
+             * "orientation" is dumped before "layout". Therefore, we store
+             * whether the orientation was horizontal or vertical in the
+             * last_split_layout. When we then encounter layout == "default",
+             * we will use the last_split_layout as layout instead. */
             char *buf = NULL;
             sasprintf(&buf, "%.*s", (int)len, val);
-            if (strcasecmp(buf, "none") == 0)
-                json_node->orientation = NO_ORIENTATION;
-            else if (strcasecmp(buf, "horizontal") == 0)
-                json_node->orientation = HORIZ;
+            if (strcasecmp(buf, "none") == 0 ||
+                strcasecmp(buf, "horizontal") == 0)
+                json_node->last_split_layout = L_SPLITH;
             else if (strcasecmp(buf, "vertical") == 0)
-                json_node->orientation = VERT;
+                json_node->last_split_layout = L_SPLITV;
             else LOG("Unhandled orientation: %s\n", buf);
             free(buf);
         } else if (strcasecmp(last_key, "border") == 0) {
@@ -171,8 +178,11 @@ static int json_string(void *ctx, const unsigned char *val, unsigned int len) {
             sasprintf(&buf, "%.*s", (int)len, val);
             if (strcasecmp(buf, "none") == 0)
                 json_node->border_style = BS_NONE;
-            else if (strcasecmp(buf, "1pixel") == 0)
-                json_node->border_style = BS_1PIXEL;
+            else if (strcasecmp(buf, "1pixel") == 0) {
+                json_node->border_style = BS_PIXEL;
+                json_node->current_border_width = 1;
+            } else if (strcasecmp(buf, "pixel") == 0)
+                json_node->border_style = BS_PIXEL;
             else if (strcasecmp(buf, "normal") == 0)
                 json_node->border_style = BS_NORMAL;
             else LOG("Unhandled \"border\": %s\n", buf);
@@ -181,7 +191,8 @@ static int json_string(void *ctx, const unsigned char *val, unsigned int len) {
             char *buf = NULL;
             sasprintf(&buf, "%.*s", (int)len, val);
             if (strcasecmp(buf, "default") == 0)
-                json_node->layout = L_DEFAULT;
+                /* This set above when we read "orientation". */
+                json_node->layout = json_node->last_split_layout;
             else if (strcasecmp(buf, "stacked") == 0)
                 json_node->layout = L_STACKED;
             else if (strcasecmp(buf, "tabbed") == 0)
@@ -190,8 +201,32 @@ static int json_string(void *ctx, const unsigned char *val, unsigned int len) {
                 json_node->layout = L_DOCKAREA;
             else if (strcasecmp(buf, "output") == 0)
                 json_node->layout = L_OUTPUT;
+            else if (strcasecmp(buf, "splith") == 0)
+                json_node->layout = L_SPLITH;
+            else if (strcasecmp(buf, "splitv") == 0)
+                json_node->layout = L_SPLITV;
             else LOG("Unhandled \"layout\": %s\n", buf);
             free(buf);
+        } else if (strcasecmp(last_key, "workspace_layout") == 0) {
+            char *buf = NULL;
+            sasprintf(&buf, "%.*s", (int)len, val);
+            if (strcasecmp(buf, "default") == 0)
+                json_node->workspace_layout = L_DEFAULT;
+            else if (strcasecmp(buf, "stacked") == 0)
+                json_node->workspace_layout = L_STACKED;
+            else if (strcasecmp(buf, "tabbed") == 0)
+                json_node->workspace_layout = L_TABBED;
+            else LOG("Unhandled \"workspace_layout\": %s\n", buf);
+            free(buf);
+        } else if (strcasecmp(last_key, "last_split_layout") == 0) {
+            char *buf = NULL;
+            sasprintf(&buf, "%.*s", (int)len, val);
+            if (strcasecmp(buf, "splith") == 0)
+                json_node->last_split_layout = L_SPLITH;
+            else if (strcasecmp(buf, "splitv") == 0)
+                json_node->last_split_layout = L_SPLITV;
+            else LOG("Unhandled \"last_splitlayout\": %s\n", buf);
+            free(buf);
         } else if (strcasecmp(last_key, "mark") == 0) {
             char *buf = NULL;
             sasprintf(&buf, "%.*s", (int)len, val);
@@ -239,6 +274,9 @@ static int json_int(void *ctx, long val) {
     if (strcasecmp(last_key, "num") == 0)
         json_node->num = val;
 
+    if (strcasecmp(last_key, "current_border_width") == 0)
+        json_node->current_border_width = val;
+
     if (!parsing_swallows && strcasecmp(last_key, "id") == 0)
         json_node->old_id = val;
 
@@ -308,11 +346,22 @@ void tree_append_json(const char *filename) {
     /* TODO: percent of other windows are not correctly fixed at the moment */
     FILE *f;
     if ((f = fopen(filename, "r")) == NULL) {
-        LOG("Cannot open file\n");
+        LOG("Cannot open file \"%s\"\n", filename);
+        return;
+    }
+    struct stat stbuf;
+    if (fstat(fileno(f), &stbuf) != 0) {
+        LOG("Cannot fstat() the file\n");
+        fclose(f);
+        return;
+    }
+    char *buf = smalloc(stbuf.st_size);
+    int n = fread(buf, 1, stbuf.st_size, f);
+    if (n != stbuf.st_size) {
+        LOG("File \"%s\" could not be read entirely, not loading.\n", filename);
+        fclose(f);
         return;
     }
-    char *buf = malloc(65535); /* TODO */
-    int n = fread(buf, 1, 65535, f);
     LOG("read %d bytes\n", n);
     yajl_gen g;
     yajl_handle hand;
